@@ -1,7 +1,6 @@
 # -*- coding=utf-8 -*-
 from __future__ import absolute_import, print_function
 
-import copy
 import operator
 import os
 import sys
@@ -11,10 +10,8 @@ from itertools import chain
 import attr
 import six
 from cached_property import cached_property
-from vistir.compat import Path, fs_str
+from ..compat import Path, fs_str
 
-from .mixins import BaseFinder, BasePath
-from .python import PythonVersion
 from ..environment import (
     ASDF_DATA_DIR,
     ASDF_INSTALLED,
@@ -22,23 +19,26 @@ from ..environment import (
     PYENV_INSTALLED,
     PYENV_ROOT,
     SHIM_PATHS,
+    get_shim_paths,
 )
 from ..exceptions import InvalidPythonVersion
 from ..utils import (
     Iterable,
     Sequence,
+    dedup,
     ensure_path,
     expand_paths,
     filter_pythons,
     is_in_path,
-    looks_like_python,
     normalize_path,
     optional_instance_of,
     parse_asdf_version_order,
     parse_pyenv_version_order,
     path_is_known_executable,
+    split_version_and_name,
     unnest,
 )
+from .mixins import BaseFinder, BasePath
 
 if MYPY_RUNNING:
     from typing import (
@@ -55,7 +55,7 @@ if MYPY_RUNNING:
         Any,
         TypeVar,
     )
-    from .python import PythonFinder
+    from .python import PythonFinder, PythonVersion
     from .windows import WindowsFinder
 
     FinderType = TypeVar("FinderType", BaseFinder, PythonFinder, WindowsFinder)
@@ -76,10 +76,9 @@ class SystemPath(object):
     path_order = attr.ib(default=attr.Factory(list))  # type: List[str]
     python_version_dict = attr.ib()  # type: DefaultDict[Tuple, List[PythonVersion]]
     only_python = attr.ib(default=False, type=bool)
-    pyenv_finder = attr.ib(
-        default=None, validator=optional_instance_of("PythonFinder")
-    )  # type: Optional[PythonFinder]
+    pyenv_finder = attr.ib(default=None)  # type: Optional[PythonFinder]
     asdf_finder = attr.ib(default=None)  # type: Optional[PythonFinder]
+    windows_finder = attr.ib(default=None)  # type: Optional[WindowsFinder]
     system = attr.ib(default=False, type=bool)
     _version_dict = attr.ib(
         default=attr.Factory(defaultdict)
@@ -91,31 +90,63 @@ class SystemPath(object):
     )  # type: Dict[str, Union[WindowsFinder, PythonFinder]]
 
     def _register_finder(self, finder_name, finder):
-        # type: (str, Union[WindowsFinder, PythonFinder]) -> None
+        # type: (str, Union[WindowsFinder, PythonFinder]) -> "SystemPath"
         if finder_name not in self.__finders:
             self.__finders[finder_name] = finder
+        return self
 
     def clear_caches(self):
         for key in ["executables", "python_executables", "version_dict", "path_entries"]:
             if key in self.__dict__:
                 del self.__dict__[key]
-        self._executables = []
-        self._python_executables = {}
-        self.python_version_dict = defaultdict(list)
-        self._version_dict = defaultdict(list)
+        for finder in list(self.__finders.keys()):
+            del self.__finders[finder]
+        self.__finders = {}
+        return attr.evolve(
+            self,
+            executables=[],
+            python_executables={},
+            python_version_dict=defaultdict(list),
+            version_dict=defaultdict(list),
+            pyenv_finder=None,
+            windows_finder=None,
+            asdf_finder=None,
+            path_order=[],
+            paths=defaultdict(PathEntry),
+        )
 
     def __del__(self):
-        self.clear_caches()
+        for key in ["executables", "python_executables", "version_dict", "path_entries"]:
+            try:
+                del self.__dict__[key]
+            except KeyError:
+                pass
+        for finder in list(self.__finders.keys()):
+            del self.__finders[finder]
+        self.__finders = {}
+        self._python_executables = {}
+        self._executables = []
+        self.python_version_dict = defaultdict(list)
+        self._version_dict = defaultdict(list)
         self.path_order = []
         self.pyenv_finder = None
         self.asdf_finder = None
         self.paths = defaultdict(PathEntry)
+        self.__finders = {}
 
     @property
     def finders(self):
         # type: () -> List[str]
         return [k for k in self.__finders.keys()]
 
+    @staticmethod
+    def check_for_pyenv():
+        return PYENV_INSTALLED or os.path.exists(normalize_path(PYENV_ROOT))
+
+    @staticmethod
+    def check_for_asdf():
+        return ASDF_INSTALLED or os.path.exists(normalize_path(ASDF_DATA_DIR))
+
     @python_version_dict.default
     def create_python_version_dict(self):
         # type: () -> DefaultDict[Tuple, List[PythonVersion]]
@@ -159,7 +190,7 @@ class SystemPath(object):
                 if entry not in self._version_dict[version] and entry.is_python:
                     self._version_dict[version].append(entry)
         for p, entry in self.python_executables.items():
-            version = entry.as_python
+            version = entry.as_python  # type: PythonVersion
             if not version:
                 continue
             if not isinstance(version, tuple):
@@ -168,35 +199,69 @@ class SystemPath(object):
                 self._version_dict[version].append(entry)
         return self._version_dict
 
-    def __attrs_post_init__(self):
-        # type: () -> None
-        #: slice in pyenv
+    def _run_setup(self):
+        # type: () -> "SystemPath"
         if not self.__class__ == SystemPath:
-            return
-        if os.name == "nt":
-            self._setup_windows()
-        if PYENV_INSTALLED:
-            self._setup_pyenv()
-        if ASDF_INSTALLED:
-            self._setup_asdf()
+            return self
+        new_instance = self
+        path_order = new_instance.path_order[:]
+        path_entries = self.paths.copy()
+        if self.global_search and "PATH" in os.environ:
+            path_order = path_order + os.environ["PATH"].split(os.pathsep)
+        path_order = list(dedup(path_order))
+        path_instances = [
+            ensure_path(p.strip('"'))
+            for p in path_order
+            if not any(
+                is_in_path(normalize_path(str(p)), normalize_path(shim))
+                for shim in SHIM_PATHS
+            )
+        ]
+        path_entries.update(
+            {
+                p.as_posix(): PathEntry.create(
+                    path=p.absolute(), is_root=True, only_python=self.only_python
+                )
+                for p in path_instances
+            }
+        )
+        new_instance = attr.evolve(
+            new_instance,
+            path_order=[p.as_posix() for p in path_instances],
+            paths=path_entries,
+        )
+        if os.name == "nt" and "windows" not in self.finders:
+            new_instance = new_instance._setup_windows()
+        #: slice in pyenv
+        if self.check_for_pyenv() and "pyenv" not in self.finders:
+            new_instance = new_instance._setup_pyenv()
+        #: slice in asdf
+        if self.check_for_asdf() and "asdf" not in self.finders:
+            new_instance = new_instance._setup_asdf()
         venv = os.environ.get("VIRTUAL_ENV")
         if os.name == "nt":
             bin_dir = "Scripts"
         else:
             bin_dir = "bin"
-        if venv and (self.system or self.global_search):
+        if venv and (new_instance.system or new_instance.global_search):
             p = ensure_path(venv)
-            self.path_order = [(p / bin_dir).as_posix()] + self.path_order
-            self.paths[p] = self.get_path(p.joinpath(bin_dir))
-        if self.system:
+            path_order = [(p / bin_dir).as_posix()] + new_instance.path_order
+            new_instance = attr.evolve(new_instance, path_order=path_order)
+            paths = new_instance.paths.copy()
+            paths[p] = new_instance.get_path(p.joinpath(bin_dir))
+            new_instance = attr.evolve(new_instance, paths=paths)
+        if new_instance.system:
             syspath = Path(sys.executable)
             syspath_bin = syspath.parent
             if syspath_bin.name != bin_dir and syspath_bin.joinpath(bin_dir).exists():
                 syspath_bin = syspath_bin / bin_dir
-            self.path_order = [syspath_bin.as_posix()] + self.path_order
-            self.paths[syspath_bin] = PathEntry.create(
+            path_order = [syspath_bin.as_posix()] + new_instance.path_order
+            paths = new_instance.paths.copy()
+            paths[syspath_bin] = PathEntry.create(
                 path=syspath_bin, is_root=True, only_python=False
             )
+            new_instance = attr.evolve(new_instance, path_order=path_order, paths=paths)
+        return new_instance
 
     def _get_last_instance(self, path):
         # type: (str) -> int
@@ -210,7 +275,7 @@ class SystemPath(object):
         return path_index
 
     def _slice_in_paths(self, start_idx, paths):
-        # type: (int, List[Path]) -> None
+        # type: (int, List[Path]) -> "SystemPath"
         before_path = []  # type: List[str]
         after_path = []  # type: List[str]
         if start_idx == 0:
@@ -220,29 +285,35 @@ class SystemPath(object):
         else:
             before_path = self.path_order[: start_idx + 1]
             after_path = self.path_order[start_idx + 2 :]
-        self.path_order = before_path + [p.as_posix() for p in paths] + after_path
+        path_order = before_path + [p.as_posix() for p in paths] + after_path
+        if path_order == self.path_order:
+            return self
+        return attr.evolve(self, path_order=path_order)
 
     def _remove_path(self, path):
-        # type: (str) -> None
+        # type: (str) -> "SystemPath"
         path_copy = [p for p in reversed(self.path_order[:])]
         new_order = []
         target = normalize_path(path)
         path_map = {normalize_path(pth): pth for pth in self.paths.keys()}
+        new_paths = self.paths.copy()
         if target in path_map:
-            del self.paths[path_map[target]]
+            del new_paths[path_map[target]]
         for current_path in path_copy:
             normalized = normalize_path(current_path)
             if normalized != target:
                 new_order.append(normalized)
         new_order = [p for p in reversed(new_order)]
-        self.path_order = new_order
+        return attr.evolve(self, path_order=new_order, paths=new_paths)
 
     def _setup_asdf(self):
-        # type: () -> None
+        # type: () -> "SystemPath"
+        if "asdf" in self.finders and self.asdf_finder is not None:
+            return self
         from .python import PythonFinder
 
         os_path = os.environ["PATH"].split(os.pathsep)
-        self.asdf_finder = PythonFinder.create(
+        asdf_finder = PythonFinder.create(
             root=ASDF_DATA_DIR,
             ignore_unsupported=True,
             sort_function=parse_asdf_version_order,
@@ -252,20 +323,25 @@ class SystemPath(object):
         try:
             asdf_index = self._get_last_instance(ASDF_DATA_DIR)
         except ValueError:
-            pyenv_index = 0 if is_in_path(next(iter(os_path), ""), PYENV_ROOT) else -1
+            asdf_index = 0 if is_in_path(next(iter(os_path), ""), ASDF_DATA_DIR) else -1
         if asdf_index is None:
             # we are in a virtualenv without global pyenv on the path, so we should
             # not write pyenv to the path here
-            return
-        root_paths = [p for p in self.asdf_finder.roots]
-        self._slice_in_paths(asdf_index, [self.asdf_finder.root])
-        self.paths[self.asdf_finder.root] = self.asdf_finder
-        self.paths.update(self.asdf_finder.roots)
-        self._remove_path(normalize_path(os.path.join(ASDF_DATA_DIR, "shims")))
-        self._register_finder("asdf", self.asdf_finder)
+            return self
+        # * These are the root paths for the finder
+        _ = [p for p in asdf_finder.roots]
+        new_instance = self._slice_in_paths(asdf_index, [asdf_finder.root])
+        paths = self.paths.copy()
+        paths[asdf_finder.root] = asdf_finder
+        paths.update(asdf_finder.roots)
+        return (
+            attr.evolve(new_instance, paths=paths, asdf_finder=asdf_finder)
+            ._remove_path(normalize_path(os.path.join(ASDF_DATA_DIR, "shims")))
+            ._register_finder("asdf", asdf_finder)
+        )
 
     def reload_finder(self, finder_name):
-        # type: (str) -> None
+        # type: (str) -> "SystemPath"
         if finder_name is None:
             raise TypeError("Must pass a string as the name of the target finder")
         finder_attr = "{0}_finder".format(finder_name)
@@ -286,19 +362,21 @@ class SystemPath(object):
             finder_name == "asdf" and not ASDF_INSTALLED
         ):
             # Don't allow loading of finders that aren't explicitly 'installed' as it were
-            pass
+            return self
         setattr(self, finder_attr, None)
         if finder_name in self.__finders:
             del self.__finders[finder_name]
-        setup_fn()
+        return setup_fn()
 
     def _setup_pyenv(self):
-        # type: () -> None
+        # type: () -> "SystemPath"
+        if "pyenv" in self.finders and self.pyenv_finder is not None:
+            return self
         from .python import PythonFinder
 
         os_path = os.environ["PATH"].split(os.pathsep)
 
-        self.pyenv_finder = PythonFinder.create(
+        pyenv_finder = PythonFinder.create(
             root=PYENV_ROOT,
             sort_function=parse_pyenv_version_order,
             version_glob_path="versions/*",
@@ -312,25 +390,37 @@ class SystemPath(object):
         if pyenv_index is None:
             # we are in a virtualenv without global pyenv on the path, so we should
             # not write pyenv to the path here
-            return
-
-        root_paths = [p for p in self.pyenv_finder.roots]
-        self._slice_in_paths(pyenv_index, [self.pyenv_finder.root])
-        self.paths[self.pyenv_finder.root] = self.pyenv_finder
-        self.paths.update(self.pyenv_finder.roots)
-        self._remove_path(os.path.join(PYENV_ROOT, "shims"))
-        self._register_finder("pyenv", self.pyenv_finder)
+            return self
+        # * These are the root paths for the finder
+        _ = [p for p in pyenv_finder.roots]
+        new_instance = self._slice_in_paths(pyenv_index, [pyenv_finder.root])
+        paths = new_instance.paths.copy()
+        paths[pyenv_finder.root] = pyenv_finder
+        paths.update(pyenv_finder.roots)
+        return (
+            attr.evolve(new_instance, paths=paths, pyenv_finder=pyenv_finder)
+            ._remove_path(os.path.join(PYENV_ROOT, "shims"))
+            ._register_finder("pyenv", pyenv_finder)
+        )
 
     def _setup_windows(self):
-        # type: () -> None
+        # type: () -> "SystemPath"
+        if "windows" in self.finders and self.windows_finder is not None:
+            return self
         from .windows import WindowsFinder
 
-        self.windows_finder = WindowsFinder.create()
-        root_paths = (p for p in self.windows_finder.paths if p.is_root)
+        windows_finder = WindowsFinder.create()
+        root_paths = (p for p in windows_finder.paths if p.is_root)
         path_addition = [p.path.as_posix() for p in root_paths]
-        self.path_order = self.path_order[:] + path_addition
-        self.paths.update({p.path: p for p in root_paths})
-        self._register_finder("windows", self.windows_finder)
+        new_path_order = self.path_order[:] + path_addition
+        new_paths = self.paths.copy()
+        new_paths.update({p.path: p for p in root_paths})
+        return attr.evolve(
+            self,
+            windows_finder=windows_finder,
+            path_order=new_path_order,
+            paths=new_paths,
+        )._register_finder("windows", windows_finder)
 
     def get_path(self, path):
         # type: (Union[str, Path]) -> PathType
@@ -350,7 +440,7 @@ class SystemPath(object):
         return _path
 
     def _get_paths(self):
-        # type: () -> Iterator
+        # type: () -> Generator[Union[PathType, WindowsFinder], None, None]
         for path in self.path_order:
             try:
                 entry = self.get_path(path)
@@ -361,7 +451,7 @@ class SystemPath(object):
 
     @cached_property
     def path_entries(self):
-        # type: () -> List[Union[PathEntry, FinderType]]
+        # type: () -> List[Union[PathType, WindowsFinder]]
         paths = list(self._get_paths())
         return paths
 
@@ -469,6 +559,7 @@ class SystemPath(object):
         dev=None,  # type: Optional[bool]
         arch=None,  # type: Optional[str]
         name=None,  # type: Optional[str]
+        sort_by_path=False,  # type: bool
     ):
         # type: (...) -> PathEntry
         """Search for a specific python version on the path.
@@ -481,29 +572,12 @@ class SystemPath(object):
         :param bool dev: Search for devreleases (default None) - prioritize releases if None
         :param str arch: Architecture to include, e.g. '64bit', defaults to None
         :param str name: The name of a python version, e.g. ``anaconda3-5.3.0``
+        :param bool sort_by_path: Whether to sort by path -- default sort is by version(default: False)
         :return: A :class:`~pythonfinder.models.PathEntry` instance matching the version requested.
         :rtype: :class:`~pythonfinder.models.PathEntry`
         """
 
-        if isinstance(major, six.string_types) and not minor and not patch:
-            # Only proceed if this is in the format "x.y.z" or similar
-            if major.isdigit() or (major.count(".") > 0 and major[0].isdigit()):
-                version = major.split(".", 2)
-                if isinstance(version, (tuple, list)):
-                    if len(version) > 3:
-                        major, minor, patch, rest = version
-                    elif len(version) == 3:
-                        major, minor, patch = version
-                    elif len(version) == 2:
-                        major, minor = version
-                    else:
-                        major = major[0]
-                else:
-                    major = major
-                    name = None
-            else:
-                name = "{0!s}".format(major)
-                major = None
+        major, minor, patch, name = split_version_and_name(major, minor, patch, name)
         sub_finder = operator.methodcaller(
             "find_python_version", major, minor, patch, pre, dev, arch, name
         )
@@ -521,6 +595,18 @@ class SystemPath(object):
             windows_finder_version = sub_finder(self.windows_finder)
             if windows_finder_version:
                 return windows_finder_version
+        if sort_by_path:
+            paths = [self.get_path(k) for k in self.path_order]
+            for path in paths:
+                found_version = sub_finder(path)
+                if found_version:
+                    return found_version
+            if alternate_sub_finder:
+                for path in paths:
+                    found_version = alternate_sub_finder(path)
+                    if found_version:
+                        return found_version
+
         ver = next(iter(self.get_pythons(sub_finder)), None)
         if not ver and alternate_sub_finder is not None:
             ver = next(iter(self.get_pythons(alternate_sub_finder)), None)
@@ -561,7 +647,19 @@ class SystemPath(object):
         if global_search:
             if "PATH" in os.environ:
                 paths = os.environ["PATH"].split(os.pathsep)
+        path_order = []  # type: List[str]
         if path:
+            path_order = [path]
+            path_instance = ensure_path(path)
+            path_entries.update(
+                {
+                    path_instance.as_posix(): PathEntry.create(
+                        path=path_instance.absolute(),
+                        is_root=True,
+                        only_python=only_python,
+                    )
+                }
+            )
             paths = [path] + paths
         paths = [p for p in paths if not any(is_in_path(p, shim) for shim in SHIM_PATHS)]
         _path_objects = [ensure_path(p.strip('"')) for p in paths]
@@ -574,23 +672,41 @@ class SystemPath(object):
                 for p in _path_objects
             }
         )
-        return cls(
+        instance = cls(
             paths=path_entries,
-            path_order=paths,
+            path_order=path_order,
             only_python=only_python,
             system=system,
             global_search=global_search,
             ignore_unsupported=ignore_unsupported,
         )
+        instance = instance._run_setup()
+        return instance
 
 
 @attr.s(slots=True)
 class PathEntry(BasePath):
-    is_root = attr.ib(default=True, type=bool)
+    is_root = attr.ib(default=True, type=bool, cmp=False)
+
+    def __lt__(self, other):
+        # type: (BasePath) -> bool
+        return self.path.as_posix() < other.path.as_posix()
+
+    def __lte__(self, other):
+        # type: (BasePath) -> bool
+        return self.path.as_posix() <= other.path.as_posix()
+
+    def __gt__(self, other):
+        # type: (BasePath) -> bool
+        return self.path.as_posix() > other.path.as_posix()
+
+    def __gte__(self, other):
+        # type: (BasePath) -> bool
+        return self.path.as_posix() >= other.path.as_posix()
 
     def __del__(self):
-        if "_children" in self.__dict__:
-            del self.__dict__["_children"]
+        if getattr(self, "_children"):
+            del self._children
         BasePath.__del__(self)
 
     def _filter_children(self):
@@ -603,8 +719,6 @@ class PathEntry(BasePath):
 
     def _gen_children(self):
         # type: () -> Iterator
-        from ..environment import get_shim_paths
-
         shim_paths = get_shim_paths()
         pass_name = self.name != self.path.name
         pass_args = {"is_root": False, "only_python": self.only_python}
@@ -630,16 +744,26 @@ class PathEntry(BasePath):
                 yield (child.as_posix(), entry)
         return
 
-    @cached_property
+    @property
     def children(self):
         # type: () -> Dict[str, PathEntry]
         children = getattr(self, "_children", {})  # type: Dict[str, PathEntry]
         if not children:
             for child_key, child_val in self._gen_children():
                 children[child_key] = child_val
-            self._children = children
+            self.children = children
         return self._children
 
+    @children.setter
+    def children(self, val):
+        # type: (Dict[str, PathEntry]) -> None
+        self._children = val
+
+    @children.deleter
+    def children(self):
+        # type: () -> None
+        del self._children
+
     @classmethod
     def create(cls, path, is_root=False, only_python=False, pythons=None, name=None):
         # type: (Union[str, Path], bool, bool, Dict[str, PythonVersion], Optional[str]) -> PathEntry
