@@ -5,14 +5,17 @@ import io
 import itertools
 import os
 import re
+import subprocess
+from collections import OrderedDict
 from fnmatch import fnmatch
+from threading import Timer
 
 import attr
 import six
-import vistir
 from packaging.version import LegacyVersion, Version
 
-from .environment import MYPY_RUNNING, PYENV_ROOT
+from .compat import Path, lru_cache, TimeoutError  # noqa
+from .environment import MYPY_RUNNING, PYENV_ROOT, SUBPROCESS_TIMEOUT
 from .exceptions import InvalidPythonVersion
 
 six.add_move(
@@ -26,22 +29,18 @@ from six.moves import Iterable  # type: ignore  # noqa  # isort:skip
 from six.moves import Sequence  # type: ignore  # noqa  # isort:skip
 # fmt: on
 
-try:
-    from functools import lru_cache
-except ImportError:
-    from backports.functools_lru_cache import lru_cache  # type: ignore  # noqa
-
 if MYPY_RUNNING:
     from typing import Any, Union, List, Callable, Set, Tuple, Dict, Optional, Iterator
     from attr.validators import _OptionalValidator  # type: ignore
     from .models.path import PathEntry
 
 
-version_re = re.compile(
+version_re_str = (
     r"(?P<major>\d+)(?:\.(?P<minor>\d+))?(?:\.(?P<patch>(?<=\.)[0-9]+))?\.?"
     r"(?:(?P<prerel>[abc]|rc|dev)(?:(?P<prerelversion>\d+(?:\.\d+)*))?)"
     r"?(?P<postdev>(\.post(?P<post>\d+))?(\.dev(?P<dev>\d+))?)?"
 )
+version_re = re.compile(version_re_str)
 
 
 PYTHON_IMPLEMENTATIONS = (
@@ -53,13 +52,22 @@ PYTHON_IMPLEMENTATIONS = (
     "miniconda",
     "stackless",
     "activepython",
+    "pyston",
     "micropython",
 )
-RE_MATCHER = re.compile(
-    r"(({0})(?:\d?(?:\.\d[cpm]{{0,3}}))?(?:-?[\d\.]+)*[^z])".format(
-        "|".join(PYTHON_IMPLEMENTATIONS)
-    )
+if os.name == "nt":
+    KNOWN_EXTS = {"exe", "py", "bat", ""}
+else:
+    KNOWN_EXTS = {"sh", "bash", "csh", "zsh", "fish", "py", ""}
+KNOWN_EXTS = KNOWN_EXTS | set(
+    filter(None, os.environ.get("PATHEXT", "").split(os.pathsep))
 )
+PY_MATCH_STR = r"((?P<implementation>{0})(?:\d?(?:\.\d[cpm]{{0,3}}))?(?:-?[\d\.]+)*[^zw])".format(
+    "|".join(PYTHON_IMPLEMENTATIONS)
+)
+EXE_MATCH_STR = r"{0}(?:\.(?P<ext>{1}))?".format(PY_MATCH_STR, "|".join(KNOWN_EXTS))
+RE_MATCHER = re.compile(r"({0}|{1})".format(version_re_str, PY_MATCH_STR))
+EXE_MATCHER = re.compile(EXE_MATCH_STR)
 RULES_BASE = [
     "*{0}",
     "*{0}?",
@@ -71,11 +79,6 @@ RULES_BASE = [
 ]
 RULES = [rule.format(impl) for impl in PYTHON_IMPLEMENTATIONS for rule in RULES_BASE]
 
-KNOWN_EXTS = {"exe", "py", "fish", "sh", ""}
-KNOWN_EXTS = KNOWN_EXTS | set(
-    filter(None, os.environ.get("PATHEXT", "").split(os.pathsep))
-)
-
 MATCH_RULES = []
 for rule in RULES:
     MATCH_RULES.extend(
@@ -87,21 +90,31 @@ for rule in RULES:
 def get_python_version(path):
     # type: (str) -> str
     """Get python version string using subprocess from a given path."""
-    version_cmd = [path, "-c", "import sys; print(sys.version.split()[0])"]
+    version_cmd = [
+        path,
+        "-c",
+        "import sys; print('.'.join([str(i) for i in sys.version_info[:3]]))",
+    ]
+    subprocess_kwargs = {
+        "env": os.environ.copy(),
+        "universal_newlines": True,
+        "stdout": subprocess.PIPE,
+        "stderr": subprocess.PIPE,
+        "shell": False,
+    }
+    c = subprocess.Popen(version_cmd, **subprocess_kwargs)
+    timer = Timer(SUBPROCESS_TIMEOUT, c.kill)
     try:
-        c = vistir.misc.run(
-            version_cmd,
-            block=True,
-            nospin=True,
-            return_object=True,
-            combine_stderr=False,
-            write_to_stdout=False,
-        )
+        out, _ = c.communicate()
+    except (SystemExit, KeyboardInterrupt, TimeoutError):
+        c.terminate()
+        out, _ = c.communicate()
+        raise
     except OSError:
         raise InvalidPythonVersion("%s is not a valid python path" % path)
-    if not c.out:
+    if not out:
         raise InvalidPythonVersion("%s is not a valid python path" % path)
-    return c.out.strip()
+    return out.strip()
 
 
 @lru_cache(maxsize=1024)
@@ -179,13 +192,13 @@ def path_is_executable(path):
 
 @lru_cache(maxsize=1024)
 def path_is_known_executable(path):
-    # type: (vistir.compat.Path) -> bool
+    # type: (Path) -> bool
     """
     Returns whether a given path is a known executable from known executable extensions
     or has the executable bit toggled.
 
     :param path: The path to the target executable.
-    :type path: :class:`~vistir.compat.Path`
+    :type path: :class:`~Path`
     :return: True if the path has chmod +x, or is a readable, known executable extension.
     :rtype: bool
     """
@@ -218,12 +231,12 @@ def looks_like_python(name):
 
 @lru_cache(maxsize=1024)
 def path_is_python(path):
-    # type: (vistir.compat.Path) -> bool
+    # type: (Path) -> bool
     """
     Determine whether the supplied path is executable and looks like a possible path to python.
 
     :param path: The path to an executable.
-    :type path: :class:`~vistir.compat.Path`
+    :type path: :class:`~Path`
     :return: Whether the provided path is an executable path to python.
     :rtype: bool
     """
@@ -231,9 +244,43 @@ def path_is_python(path):
     return path_is_executable(path) and looks_like_python(path.name)
 
 
+@lru_cache(maxsize=1024)
+def guess_company(path):
+    # type: (str) -> Optional[str]
+    """Given a path to python, guess the company who created it
+
+    :param str path: The path to guess about
+    :return: The guessed company
+    :rtype: Optional[str]
+    """
+    non_core_pythons = [impl for impl in PYTHON_IMPLEMENTATIONS if impl != "python"]
+    return next(
+        iter(impl for impl in non_core_pythons if impl in path.lower()), "PythonCore"
+    )
+
+
+@lru_cache(maxsize=1024)
+def path_is_pythoncore(path):
+    # type: (str) -> bool
+    """Given a path, determine whether it appears to be pythoncore.
+
+    Does not verify whether the path is in fact a path to python, but simply
+    does an exclusionary check on the possible known python implementations
+    to see if their names are present in the path (fairly dumb check).
+
+    :param str path: The path to check
+    :return: Whether that path is a PythonCore path or not
+    :rtype: bool
+    """
+    company = guess_company(path)
+    if company:
+        return company == "PythonCore"
+    return False
+
+
 @lru_cache(maxsize=1024)
 def ensure_path(path):
-    # type: (Union[vistir.compat.Path, str]) -> vistir.compat.Path
+    # type: (Union[Path, str]) -> Path
     """
     Given a path (either a string or a Path object), expand variables and return a Path object.
 
@@ -243,9 +290,9 @@ def ensure_path(path):
     :rtype: :class:`~pathlib.Path`
     """
 
-    if isinstance(path, vistir.compat.Path):
+    if isinstance(path, Path):
         return path
-    path = vistir.compat.Path(os.path.expandvars(path))
+    path = Path(os.path.expandvars(path))
     return path.absolute()
 
 
@@ -268,10 +315,10 @@ def normalize_path(path):
 
 @lru_cache(maxsize=1024)
 def filter_pythons(path):
-    # type: (Union[str, vistir.compat.Path]) -> Iterable
+    # type: (Union[str, Path]) -> Iterable
     """Return all valid pythons in a given path"""
-    if not isinstance(path, vistir.compat.Path):
-        path = vistir.compat.Path(str(path))
+    if not isinstance(path, Path):
+        path = Path(str(path))
     if not path.is_dir():
         return path if path_is_python(path) else None
     return filter(path_is_python, path.iterdir())
@@ -326,6 +373,35 @@ def parse_asdf_version_order(filename=".tool-versions"):
     return []
 
 
+def split_version_and_name(
+    major=None,  # type: Optional[Union[str, int]]
+    minor=None,  # type: Optional[Union[str, int]]
+    patch=None,  # type: Optional[Union[str, int]]
+    name=None,  # type: Optional[str]
+):
+    # type: (...) -> Tuple[Optional[Union[str, int]], Optional[Union[str, int]], Optional[Union[str, int]], Optional[str]]  # noqa
+    if isinstance(major, six.string_types) and not minor and not patch:
+        # Only proceed if this is in the format "x.y.z" or similar
+        if major.isdigit() or (major.count(".") > 0 and major[0].isdigit()):
+            version = major.split(".", 2)
+            if isinstance(version, (tuple, list)):
+                if len(version) > 3:
+                    major, minor, patch, _ = version
+                elif len(version) == 3:
+                    major, minor, patch = version
+                elif len(version) == 2:
+                    major, minor = version
+                else:
+                    major = major[0]
+            else:
+                major = major
+                name = None
+        else:
+            name = "{0!s}".format(major)
+            major = None
+    return (major, minor, patch, name)
+
+
 # TODO: Reimplement in vistir
 def is_in_path(path, parent):
     return normalize_path(str(path)).startswith(normalize_path(str(parent)))
@@ -363,3 +439,11 @@ def expand_paths(path, only_python=True):
     else:
         if path is not None and path.is_python and path.as_python is not None:
             yield path
+
+
+def dedup(iterable):
+    # type: (Iterable) -> Iterable
+    """Deduplicate an iterable object like iter(set(iterable)) but
+    order-reserved.
+    """
+    return iter(OrderedDict.fromkeys(iterable))
